#
# Copyright (c) 2023, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Union, Any, Optional, List, Callable

import oser
from htf.hart.prompt import prompt
from prompt_toolkit.document import Document  # type: ignore
from prompt_toolkit.completion import WordCompleter  # type: ignore
from prompt_toolkit.validation import Validator, ValidationError  # type: ignore
import re
import logging
import copy

from oser.types import OserElementType

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


class NumberValidator(Validator):  # type: ignore
    """Validator for number inputs"""

    def validate(self, document: Document) -> None:  # type: ignore
        """
        Validates that the given document text only contains digits.

         Args:
             document: document to validate

         Raises:
             ValidationError: If the document contains non-digit characters
        """
        text = document.text

        if text and not text.isdigit():
            i = 0

            # Get index of fist non-numeric character.
            # We want to move the cursor here.
            for i, c in enumerate(text):  # pragma: no cover
                if not c.isdigit():
                    break

            raise ValidationError(message="This input contains non-numeric characters", cursor_position=i)


class FloatValidator(Validator):  # type: ignore
    """Validator for float inputs"""

    def validate(self, document: Document) -> None:  # type: ignore
        """
        Validates that the given document text only contains a float.

         Args:
             document: document to validate

         Raises:
             ValidationError: If the document text doesn't match the float regex.
        """
        text = document.text

        if text and not (re.match(r"^\d*(\.\d+){0,1}$", text) or re.match(r"^[nN][aA][nN]$", text)):
            i = 0

            # Get index of fist non numeric character.
            # We want to move the cursor here.
            for i, c in enumerate(text):  # pragma: no cover
                if c != "." and not c.isdigit():
                    break

            raise ValidationError(message="This input contains non-numeric characters", cursor_position=i)


class BoolValidator(Validator):  # type: ignore
    """Validator for float inputs"""

    def validate(self, document: Document) -> None:  # type: ignore
        """
        Validates that the given document text only contains a boolean (or int).

         Args:
             document: document to validate

         Raises:
             ValidationError: If the document text is nether boolean nor int
        """
        text = document.text

        if text.lower() not in ["true", "false"]:
            if not text.isdigit():
                raise ValidationError(message="Input has to be boolean", cursor_position=0)


def _bool_cast(s: str) -> Union[int, bool]:
    s = s.lower()
    if s == "true":
        return True
    elif s == "false":
        return False
    else:
        return int(s)


def _encode_string_cast(s: str) -> bytes:
    return "{}".format(s).encode()


def _generator_enum_validator(instance: OserElementType) -> Callable[..., Any]:
    prototype_cast = OserPrompt()._get_cast(instance._prototype)

    def _enum_validator(instance: OserElementType) -> Callable[..., Any]:
        try:
            instance = prototype_cast(instance)
        except Exception:
            pass
        return instance  # type: ignore

    return _enum_validator


class OserPrompt:
    _structs = (oser.ByteStruct,)
    _enums = (oser.Enum,)
    _strings = (oser.String, oser.Data)
    _floats = (
        oser.BFloat,
        oser.BDouble,
        oser.LFloat,
        oser.LDouble,
    )
    _bools = (oser.Flag, oser.PaddingFlag)
    _conditionals = (oser.If, oser.IfElse, oser.Switch)

    def _prompt(  # type: ignore
        self, message: str, completer: Optional[WordCompleter] = None, validator: Optional[Validator] = None
    ) -> str:
        return prompt(message=message, completer=completer, validator=validator)  # type: ignore

    def _get_class_name(self, instance: OserElementType) -> str:
        # defaults
        classname = instance.__class__.__name__

        if isinstance(instance, self._enums):
            classname += " of " + instance._prototype.__class__.__name__

        return classname  # type: ignore

    def _get_choices(self, instance: OserElementType) -> List[str]:
        # defaults
        choices: List[str] = []

        if isinstance(instance, self._enums):
            for key, value in sorted(instance._values.items()):
                choices += ["{}".format(key), "{}".format(value)]

        return choices

    def _get_validator(self, instance: OserElementType) -> Optional[Validator]:  # type: ignore
        # defaults
        validator: Optional[Validator] = NumberValidator()  # type: ignore

        # Enum
        if isinstance(instance, self._enums):
            validator = None

        # Floats
        if isinstance(instance, self._floats):
            validator = FloatValidator()

        # Strings
        if isinstance(instance, self._strings):
            validator = None

        # Bools
        if isinstance(instance, self._bools):
            validator = BoolValidator()

        return validator

    def _get_cast(self, instance: OserElementType) -> Callable[..., Any]:
        cast: Any = int

        # Enum
        if isinstance(instance, self._enums):
            cast = _generator_enum_validator(instance)

        # Floats
        if isinstance(instance, self._floats):
            cast = float

        # Strings
        if isinstance(instance, self._strings):
            cast = _encode_string_cast

        # Bools
        if isinstance(instance, self._bools):
            cast = _bool_cast

        return cast  # type: ignore

    def _get_wordcompleter(self, choices: List[str]) -> WordCompleter:  # type: ignore
        return WordCompleter(choices)

    def fill_oser_instance(
        self, instance: OserElementType, prefix: str = "", inputs: Optional[List[Any]] = None
    ) -> None:
        inputs = copy.deepcopy(inputs)
        self._fill_oser_instance(instance, prefix=prefix, inputs=inputs)

    def _fill_oser_instance(
        self, instance: OserElementType, prefix: str = "", inputs: Optional[List[Any]] = None
    ) -> None:
        # print("_fill", container)
        if not prefix:
            prefix = instance.__class__.__name__ + "()."
        if hasattr(instance, "_items"):
            for name, instance in instance._items():
                next_prefix = prefix
                if not isinstance(instance, self._conditionals):
                    next_prefix = prefix + name + "."
                self._fill_oser_instance(instance, prefix=next_prefix, inputs=inputs)
        else:
            # fill a field that is not a struct
            name = prefix[:-1]
            self._fill_oser_field(name, instance, "", inputs)

    def _fill_oser_field(
        self, name: str, instance: OserElementType, prefix: str = "", inputs: Optional[List[Any]] = None
    ) -> None:
        input_done = False
        while input_done is False:
            classname = self._get_class_name(instance)
            choices = self._get_choices(instance)
            validator = self._get_validator(instance)
            cast = self._get_cast(instance)
            # print("choices:", choices)

            if inputs:
                value = inputs.pop(0)
                # print("using value from inputs:", value)
                if choices:
                    assert value in choices + [""]
            else:
                completer = self._get_wordcompleter(choices)
                value = self._prompt(
                    "Enter value for %s (%s) [%s]> " % (prefix + name, classname, repr(instance.get())),
                    completer=completer,
                    validator=validator,
                )
            # print("value:", value, "type:", type(value))
            if value != "":
                value = cast(value)
                # print("casted value:", value, "type:", type(value))
                try:
                    instance.set(value)
                    input_done = True
                except Exception as e:
                    print("Could not set value:", str(e))
                    raise
            else:
                # print("unchanged")
                # no input
                input_done = True


def fill_oser_instance(instance: OserElementType, prefix: str = "", inputs: Optional[List[Any]] = None) -> None:
    """
    Fill an oser instance interactively.

    Args:
        instance: an oser instance to be filled.
        prefix="" (str): the prefix that is printed on interactive prompt.
        inputs=None (list of inputs): a list of inputs to be used instead of
            user inputs.
    """
    oser_prompt = OserPrompt()
    oser_prompt.fill_oser_instance(instance=instance, prefix=prefix, inputs=inputs)
